-- The Game of Life client
-- Author: Sean Seefried

module Main(module Main, module GameOfLife) where

import Graphics.UI.WX
import Graphics.UI.WXCore
import System.Exit
import Data.IORef
import Control.Concurrent.MVar
import Control.Exception

import GameOfLife


-- Constants
cellSize = 10 -- in pixels
defaultInterval = 500 -- Default interval between generations (in milliseconds)


slowInterval = 1000
mediumInterval = 300
fastInterval = 100

intervals = [slowInterval, mediumInterval, fastInterval]

defaultWorldStr = "...............\n" ++
		  "...............\n" ++
		  "...............\n" ++
		  "...............\n" ++
		  "...............\n"

--------------------
main :: IO ()
main = start lifeGUI

lifeGUI :: IO ()
lifeGUI = do
  f <- frame [ text := "Game Of Life", resizeable := False
	     , visible := False ]
  file <- menuPane [ text := "&File" ]


  hlp <- menuHelp [ text := "Help" ]
  about <- menuAbout hlp [help := "About Game Of Life"]


  worldRef <- newIORef undefined

  memDC <- memoryDCCreate



  -- The evolution timer
  evolveTimer <- timer f [on command := return ()]

  -- A boolean value to hold whether the simulation is running or not
  runningRef <- newIORef False

  --  Create a radio box to modify the interval
  ivalRadioBox <- radioBox f Horizontal ["Slow", "Medium", "Fast"] 
		  [ text := "Speed"]
  set ivalRadioBox [ on select := 
		       do { sel <- get ivalRadioBox selection
			  ; set evolveTimer [ interval := intervals !! sel
					    , enabled := True ]
			  }]
  evolvingRef <- newMVar () -- Is the board *not* evolving?

  genLabelText <- staticText f [ text := "Generation: " ]
  genCountText <- staticText f [ text := "0" ]
  genCountRef <- newIORef 0 -- number of generations

  startButton <- button f [ text := "Start" ]
  stopButton <- button f [ text := "Stop", enabled := False] 

  set startButton 
	  [ on command := do
            set startButton [ enabled := False ]
            set stopButton  [ enabled := True ]
            writeIORef runningRef True
	    sel <- get ivalRadioBox selection
            set evolveTimer 
	     [ interval := (intervals !! sel)
	     , enabled := True
             , on command := 
	       do { freeToEvolve <- tryTakeMVar evolvingRef
		  ; case freeToEvolve of 
		    Just () -> 
		     do { evolveAndDisplay f memDC genCountRef 
                          genCountText worldRef 
 			; repaint f 
		          -- Allow evolution to occur again once the
			  -- the system has become idle again.
			; set f [ on idle :=
                 		  do { putMVar evolvingRef ()
				       -- must make sure we reset the idle
				       -- event to do nothing once it 
				       -- has put the MVar back.
				     ; set f [ on idle :=  return True ]
				     ; return True}
				]
			}
		    Nothing -> do return ()
		  }
	     ]
	  ]
		    
  let stopEvolution = do { writeIORef runningRef False
		         ; set startButton [ enabled := True]
			 ; set stopButton [ enabled := False ]
			 ; set evolveTimer [ enabled := False
					   , on command := return ()]}

  set stopButton [on command :=  stopEvolution ]
  bmpRef <- newIORef Nothing

  let lifeLayoutFun sbmp =  
          margin 10 $ column 5 ([ bitmapLayout sbmp
					 , radioBoxLayout
  					 , genLayout
					 , buttonLayout ])
      bitmapLayout sbmp = row 5 [hglue, widget sbmp, hglue ]
      radioBoxLayout = row 5 [hglue, widget ivalRadioBox, hglue]
      buttonLayout = row 5  
		     [ hglue, widget startButton, hspace 5, widget stopButton, 
		       hglue]
      genLayout = row 5 [hglue, widget genLabelText, widget genCountText, hglue ]

  menuItem file 
	      [ text := "&Open Life file"
	      , on command := 
                 do { stopEvolution
		    ; loadAndInitialise f memDC worldRef lifeLayoutFun
		         bmpRef genCountText genCountRef
		    }
	      ]
  menuQuit file [ help := "Quit", on command := close f ]


  -- Set the frame's attributes
  set f [ menubar := [file, hlp]
	, on closing := exitWith ExitSuccess
	, on (menu about) := infoDialog f "About Game Of Life" 
	  "Written by Sean Seefried"
	]
  -- initialise the board
  initialise f memDC worldRef lifeLayoutFun bmpRef
	      genCountText genCountRef defaultWorldStr
  return ()
			


-- | Evolves the world by one stemp and displays it.
evolveAndDisplay :: Frame () 
		 -> DC a  
		 -> IORef Int 
		 -> StaticText () 
		 -> IORef World 
		 ->  IO ()
evolveAndDisplay f dc genCountRef genCountText worldRef = do
  world@(_, board) <- readIORef worldRef
  gens <- readIORef genCountRef
  world'@((w,h), board') <- 
    Control.Exception.catch (evaluate (evolve world)) 
      (\_ -> do { infoDialog f "Error" 
		    ("Function 'evolve' in module GameOfLife failed." ++
		     "Quitting.\n")
		; exitWith (ExitFailure 1)
		})
  drawDrawList dc h (createDrawList board' board)
  writeIORef worldRef world'
  writeIORef genCountRef (gens + 1)
  set genCountText [ text := show (gens + 1) ]

-- Draws the board on a drawing context
-- The parameter h is the height of the board (in cells) minus 1.
drawBoard :: DC a -> Int -> Board -> IO ()
drawBoard dc h board =  drawBoard' 0 board
 where
  drawBoard'  _ [] = return ()
  drawBoard'  x (col:cols) = do 
    drawCol (x,0) col
    drawBoard' (x+1) cols
  drawCol _ [] = return ()
  drawCol (x,y) (b:rest) = do
    drawCell dc h (x,y) b
    drawCol (x, y+1) rest


-- | A DrawList is a list of cells that need to be drawn this frame.
--   Generated by comparing current board with board of last frame and 
--   seeing what is different.
type DrawList = [(Int,Int, Bool)]

-- | Creates a DrawList by comparing the first board with the second board.
--   The first board is the newer board. i.e. to be drawn this frame
--   Precondition: Boards must be of exactly the same size.
createDrawList :: Board -> Board -> DrawList
createDrawList = createDrawList' (0,0) 
 where
   createDrawList' (x,y) [] [] = []
   createDrawList' (x,y) ([]:c1ss) ([]:c2ss) =
     createDrawList' (x+1, 0) c1ss c2ss
   createDrawList' (x,y) ((c1:c1s):c1ss) ((c2:c2s):c2ss) 
     | c1 == c2 = createDrawList' (x, y+1) (c1s:c1ss) (c2s:c2ss)
     | otherwise = (x,y,c1):createDrawList' (x, y+1) (c1s:c1ss) (c2s:c2ss)
   createDrawList' _ _ _ = error "Boards are not of same size"

-- | drawDrawList draws a DrawList on a Device Context.
drawDrawList :: DC a -> Int -> DrawList -> IO ()
drawDrawList dc h [] = return ()
drawDrawList dc h ((x,y,c):rest) = do
  drawCell dc h (x,y) c
  drawDrawList dc h rest



drawCell :: DC a -> Int -> (Int,Int) -> Cell -> IO ()
drawCell dc h (x,y) c =
  let x' = x*cellSize
      y' = (h-y)*cellSize
      col = if c then red else white
  in do drawRect dc (Rect x' y' (cellSize+1) (cellSize+1))
		     [color := black, bgcolor := col]

-- | loadAndInitialise presents a dialogue box to select a world file.  
--   This file is read and the contents passed to 'initialise'.
--   FIXME: Will die if world file is not valid
loadAndInitialise f dc worldRef layoutFun bmpRef genCountText genCountRef = do
  mFilePath <- fileOpenDialog f True True "Select Game file"
	         [("Text file", ["*.txt"])] "" ""
  case mFilePath of
    Just filePath -> do
      worldStr <- readFile filePath
      initialise f dc worldRef layoutFun bmpRef genCountText 
		 genCountRef worldStr 
    Nothing -> return ()


-- | initialise calls 'initWorld' on the given world string and
--   then a) creates a bitmap representation of it
--        b) creates the correct layout for the window
--        c) draws the board
--        d) repaints window
initialise f memDC worldRef layoutFun bmpRef 
	   genCountText genCountRef worldStr = do

  loaded <- Control.Exception.catch 
	     (do {evaluate (initWorld worldStr); return True}) 
	     (\_ -> return False)
  if loaded 
   then do
    set f [ visible := False]
    let world@((x,y), board) = initWorld worldStr
        (w,h) = ((x+1)*cellSize+1, (y+1)*cellSize+1)
    writeIORef worldRef world
    -- Load bitmaps, delete them and create new ones.
    mBmpRefs <- readIORef bmpRef
    case mBmpRefs of 
      Just (bmp, sbmp) -> do
         bitmapDelete bmp
         staticBitmapDelete sbmp
      Nothing -> return ()
    bmp' <- bitmapCreateEmpty (Size w h) (-1)
    sbmp' <- staticBitmapCreate f 0 bmp' rectNull 0
    writeIORef bmpRef (Just (bmp', sbmp'))

    memoryDCSelectObject memDC bmp'
    dcClear memDC -- clears the bitmap to white

    -- reset generation counter
    set genCountText [ text := "0" ]
    writeIORef genCountRef 0
    set f [ layout := layoutFun sbmp' ]
    drawBoard memDC y board
    windowReLayoutMinimal f
    set f [ visible := True ]
   else
    infoDialog f "Error" 
     ("Function 'initWorld' in module GameOfLife failed.\n"  ++
      "This may be because initWorld is not correct or because\n" ++
      "the world file is not valid.\n")


-------


